/***********************************************************************************************************************
 * Copyright [2020-2022] Renesas Electronics Corporation and/or its affiliates.  All Rights Reserved.
 *
 * This software and documentation are supplied by Renesas Electronics America Inc. and may only be used with products
 * of Renesas Electronics Corp. and its affiliates ("Renesas").  No other uses are authorized.  Renesas products are
 * sold pursuant to Renesas terms and conditions of sale.  Purchasers are solely responsible for the selection and use
 * of Renesas products and Renesas assumes no liability.  No license, express or implied, to any intellectual property
 * right is granted by Renesas. This software is protected under all applicable laws, including copyright laws. Renesas
 * reserves the right to change or discontinue this software and/or this documentation. THE SOFTWARE AND DOCUMENTATION
 * IS DELIVERED TO YOU "AS IS," AND RENESAS MAKES NO REPRESENTATIONS OR WARRANTIES, AND TO THE FULLEST EXTENT
 * PERMISSIBLE UNDER APPLICABLE LAW, DISCLAIMS ALL WARRANTIES, WHETHER EXPLICITLY OR IMPLICITLY, INCLUDING WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND NONINFRINGEMENT, WITH RESPECT TO THE SOFTWARE OR
 * DOCUMENTATION.  RENESAS SHALL HAVE NO LIABILITY ARISING OUT OF ANY SECURITY VULNERABILITY OR BREACH.  TO THE MAXIMUM
 * EXTENT PERMITTED BY LAW, IN NO EVENT WILL RENESAS BE LIABLE TO YOU IN CONNECTION WITH THE SOFTWARE OR DOCUMENTATION
 * (OR ANY PERSON OR ENTITY CLAIMING RIGHTS DERIVED FROM YOU) FOR ANY LOSS, DAMAGES, OR CLAIMS WHATSOEVER, INCLUDING,
 * WITHOUT LIMITATION, ANY DIRECT, CONSEQUENTIAL, SPECIAL, INDIRECT, PUNITIVE, OR INCIDENTAL DAMAGES; ANY LOST PROFITS,
 * OTHER ECONOMIC DAMAGE, PROPERTY DAMAGE, OR PERSONAL INJURY; AND EVEN IF RENESAS HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH LOSS, DAMAGES, CLAIMS OR COSTS.
 **********************************************************************************************************************/

#include "blinky_thread.h"
#include "bsl_isl29035.h"
#include "bsl_hs3003.h"
#include "my_da16200_AT.h"
#include <stdio.h>
#include "string.h"

extern bsp_leds_t g_bsp_leds;
uint8_t send_log[100] = {0};

uint8_t light_value_cmd[10] = {0x5A,0xA5,0x08,0x82,0x12,0x00,0x00,0x00,0x00,0x00};
uint8_t light_process_cmd[] = {0x5A,0xA5, 0x11, 0x82, 0x14,0x00, 0x00,0x04, 0x00,0x01, 0x03,0x06, 0x00,0x2C, 0x03,0xC6, 0x00,0x3A, 0x4B,0x5A};
uint8_t temp_humi_cmd[] = {0x5A,0xA5,14,0x82,0x10,0x00,'T',':','0','0','C',' ','H',':','0','0','%'};
extern uint8_t decting_flag;
uint8_t udp_send_data[4] = {0x01,0x00,0x00,0x00};   //帧头0x01，温度，湿度，亮度

void dwin_light_send(uint32_t light)
{
    if(decting_flag == 1)
        return;
    if(light > 1000)
        light = 1000;
    light /= 10;

    if(light >= 100)
    {
        light_value_cmd[2] = 0x08;
        light_value_cmd[6] = '1';
        light_value_cmd[7] = '0';
        light_value_cmd[8] = '0';
        light_value_cmd[9] = '\0';
    }
    else if(light >= 10)
    {
        light_value_cmd[2] = 0x07;
        light_value_cmd[6] = (uint8_t)light / 10 + '0';
        light_value_cmd[7] = (uint8_t)light % 10 + '0';
        light_value_cmd[8] = '\0';
    }
    else
    {
        light_value_cmd[2] = 0x05;
        light_value_cmd[6] = (uint8_t)light % 10 + '0';
        light_value_cmd[7] = '\0';
    }
    udp_send_data[3] = (uint8_t)(light);
    light_process_cmd[15] = (uint8_t)(0x06 + light * 1.92);
    R_SCI_UART_Write(&g_uart2_ctrl, light_value_cmd, sizeof(light_value_cmd));
    vTaskDelay(10);
    R_SCI_UART_Write(&g_uart2_ctrl, light_process_cmd, sizeof(light_process_cmd));
    vTaskDelay(10);
}

void dwin_temp_humi_send(uint32_t temp, uint32_t humi)
{
    if(decting_flag == 1)
        return;
    //sprintf(&temp_humi_cmd[6], "T:%2dC H:%2d%", temp, humi);
    udp_send_data[1] = (uint8_t)temp;
    udp_send_data[2] = (uint8_t)humi;
    temp_humi_cmd[8] = (uint8_t)(temp/10 + '0');
    temp_humi_cmd[9] = (uint8_t)(temp%10 + '0');
    temp_humi_cmd[14] = (uint8_t)(humi/10 + '0');
    temp_humi_cmd[15] = (uint8_t)(humi%10 + '0');
    R_SCI_UART_Write(&g_uart2_ctrl, temp_humi_cmd, sizeof(temp_humi_cmd));
    vTaskDelay(10);
    my_da16200_send_data(udp_send_data, sizeof(udp_send_data));
}

uint32_t lux;
uint32_t temp, humi;

/* Blinky Thread entry function */
void blinky_thread_entry (void * pvParameters)
{
    FSP_PARAMETER_NOT_USED(pvParameters);
    uint8_t second = 0;
    /* LED type structure */


    s_Hs300xDataType humi_temp;
    bsp_leds_t leds = g_bsp_leds;

    /* If this board has no LEDs then trap here */
    if (0 == leds.led_count)
    {
        while (1)
        {
            ;                          // There are no LEDs on this board
        }
    }

    /* Holds level to set for pins */
    bsp_io_level_t pin_level = BSP_IO_LEVEL_LOW;
    Isl29035_Init();

    while (1)
    {

        if(second >= 10)
        {
            second = 0;
            humi_temp = Hs300x_ReadData();
            temp = (uint32_t)humi_temp.temp;
            humi = (uint32_t)humi_temp.humi;
            //sprintf((char *)send_log, "lux:%lu temp:%d humo:%d\r\n", lux, temp, humi);
            //R_SCI_UART_Write(&g_uart9_ctrl, send_log, strlen(send_log));
            /* Enable access to the PFS registers. If using r_ioport module then register protection is automatically
             * handled. This code uses BSP IO functions to show how it is used.
             */
            dwin_temp_humi_send(temp, humi);
            R_BSP_PinAccessEnable();

            /* Update all board LEDs */
            for (uint32_t i = 0; i < leds.led_count; i++)
            {
                /* Get pin to toggle */
                uint32_t pin = leds.p_leds[i];

                /* Write to this pin */
                R_BSP_PinWrite((bsp_io_port_pin_t) pin, pin_level);
            }

            /* Protect PFS registers */
            R_BSP_PinAccessDisable();

            /* Toggle level for next write */
            if (BSP_IO_LEVEL_LOW == pin_level)
            {
                pin_level = BSP_IO_LEVEL_HIGH;
            }
            else
            {
                pin_level = BSP_IO_LEVEL_LOW;
            }
        }
        second++;

        lux = Isl29035_ReadData();
        dwin_light_send(lux);

        vTaskDelay(90);
    }
}
